import json
import os
import time
from contextlib import contextmanager
from pathlib import Path
from uuid import uuid4

import pytest
from databricks.sql import connect
from databricks.sql.client import Connection as DeltaTableConnection
from databricks.sql.client import Cursor as DeltaTableCursor
from pydantic import BaseModel, Secret, SecretStr
from pytest_mock import MockerFixture

from test.integration.connectors.utils.constants import DESTINATION_TAG, SQL_TAG, env_setup_path
from test.integration.utils import requires_env
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connectors.sql.databricks_delta_tables import (
    CONNECTOR_TYPE,
    DatabricksDeltaTablesAccessConfig,
    DatabricksDeltaTablesConnectionConfig,
    DatabricksDeltaTablesUploader,
    DatabricksDeltaTablesUploaderConfig,
    DatabricksDeltaTablesUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

CATALOG = "utic-dev-tech-fixtures"


class EnvData(BaseModel):
    server_hostname: str
    http_path: str
    access_token: SecretStr


def get_env_data() -> EnvData:
    return EnvData(
        server_hostname=os.environ["DATABRICKS_SERVER_HOSTNAME"],
        http_path=os.environ["DATABRICKS_HTTP_PATH"],
        access_token=os.environ["DATABRICKS_ACCESS_TOKEN"],
    )


def get_destination_schema(new_table_name: str) -> str:
    p = Path(env_setup_path / "sql" / "databricks_delta_tables" / "destination" / "schema.sql")
    with p.open() as f:
        data_lines = f.readlines()
    data_lines[0] = data_lines[0].replace("elements", new_table_name)
    data = "".join([line.strip() for line in data_lines])
    return data


@contextmanager
def get_connection() -> DeltaTableConnection:
    env_data = get_env_data()
    with connect(
        server_hostname=env_data.server_hostname,
        http_path=env_data.http_path,
        access_token=env_data.access_token.get_secret_value(),
    ) as connection:
        yield connection


@contextmanager
def get_cursor() -> DeltaTableCursor:
    with get_connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute(f"USE CATALOG '{CATALOG}'")
            yield cursor


@pytest.fixture
def destination_table() -> str:
    random_id = str(uuid4())[:8]
    table_name = f"elements_{random_id}"
    destination_schema = get_destination_schema(new_table_name=table_name)
    with get_cursor() as cursor:
        logger.info(f"creating table: {table_name}")
        cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
        cursor.execute(destination_schema)

    yield table_name
    with get_cursor() as cursor:
        logger.info(f"dropping table: {table_name}")
        cursor.execute(f"DROP TABLE IF EXISTS {table_name}")


def validate_destination(expected_num_elements: int, table_name: str, retries=30, interval=1):
    with get_cursor() as cursor:
        for i in range(retries):
            cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
            count = cursor.fetchone()[0]
            if count == expected_num_elements:
                break
            logger.info(f"retry attempt {i}: expected {expected_num_elements} != count {count}")
            time.sleep(interval)
        assert (
            count == expected_num_elements
        ), f"dest check failed: got {count}, expected {expected_num_elements}"


@pytest.mark.asyncio
@pytest.mark.skip("Resources take too long to spin up to run in CI")
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
@requires_env("DATABRICKS_SERVER_HOSTNAME", "DATABRICKS_HTTP_PATH", "DATABRICKS_ACCESS_TOKEN")
async def test_databricks_delta_tables_destination(
    upload_file: Path, temp_dir: Path, destination_table: str
):
    env_data = get_env_data()
    mock_file_data = FileData(
        identifier="mock file data",
        connector_type=CONNECTOR_TYPE,
        source_identifiers=SourceIdentifiers(filename=upload_file.name, fullpath=upload_file.name),
    )
    stager = DatabricksDeltaTablesUploadStager()
    staged_path = stager.run(
        elements_filepath=upload_file,
        file_data=mock_file_data,
        output_dir=temp_dir,
        output_filename=upload_file.name,
    )

    assert staged_path.suffix == upload_file.suffix

    uploader = DatabricksDeltaTablesUploader(
        connection_config=DatabricksDeltaTablesConnectionConfig(
            access_config=DatabricksDeltaTablesAccessConfig(
                token=env_data.access_token.get_secret_value()
            ),
            http_path=env_data.http_path,
            server_hostname=env_data.server_hostname,
        ),
        upload_config=DatabricksDeltaTablesUploaderConfig(
            catalog=CATALOG, database="default", table_name=destination_table
        ),
    )
    with staged_path.open("r") as f:
        staged_data = json.load(f)
    expected_num_elements = len(staged_data)
    uploader.precheck()
    uploader.run(path=staged_path, file_data=mock_file_data)
    validate_destination(expected_num_elements=expected_num_elements, table_name=destination_table)


def test_get_credentials_provider_with_client_id_and_secret(mocker: MockerFixture):
    access_config = DatabricksDeltaTablesAccessConfig(
        client_id="test_client_id", client_secret="test_client_secret"
    )
    connection_config = DatabricksDeltaTablesConnectionConfig(
        access_config=Secret(access_config),
        server_hostname="test_server_hostname",
        http_path="test_http_path",
    )

    credentials_provider = connection_config.get_credentials_provider()
    assert credentials_provider is not False
    assert type(credentials_provider).__name__ == "function"


def test_get_credentials_provider_with_token(mocker: MockerFixture):
    access_config = DatabricksDeltaTablesAccessConfig(token="test_token")
    connection_config = DatabricksDeltaTablesConnectionConfig(
        access_config=Secret(access_config),
        server_hostname="test_server_hostname",
        http_path="test_http_path",
    )

    credentials_provider = connection_config.get_credentials_provider()
    assert credentials_provider is False



import os
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field

from unstructured_ingest.utils.data_prep import get_data_df, write_data
from unstructured_ingest.v2.interfaces import Uploader, UploaderConfig
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.databricks.volumes import DatabricksPathMixin
from unstructured_ingest.v2.processes.connectors.sql.databricks_delta_tables import (
    DatabricksDeltaTablesConnectionConfig,
    DatabricksDeltaTablesUploadStager,
    DatabricksDeltaTablesUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData

CONNECTOR_TYPE = "databricks_volume_delta_tables"

if TYPE_CHECKING:
    from pandas import DataFrame


class DatabricksVolumeDeltaTableUploaderConfig(UploaderConfig, DatabricksPathMixin):
    database: str = Field(description="Database name", default="default")
    table_name: str = Field(description="Table name")


@dataclass
class DatabricksVolumeDeltaTableStager(DatabricksDeltaTablesUploadStager):
    def write_output(self, output_path: Path, data: list[dict]) -> Path:
        # To avoid new line issues when migrating from volumes into delta tables, omit indenting
        # and always write it as a json file
        final_output_path = output_path.with_suffix(".json")
        write_data(path=final_output_path, data=data, indent=None)
        return final_output_path


@dataclass
class DatabricksVolumeDeltaTableUploader(Uploader):
    connection_config: DatabricksDeltaTablesConnectionConfig
    upload_config: DatabricksVolumeDeltaTableUploaderConfig
    connector_type: str = CONNECTOR_TYPE
    _columns: Optional[dict[str, str]] = None

    def precheck(self) -> None:
        with self.connection_config.get_cursor() as cursor:
            cursor.execute("SHOW CATALOGS")
            catalogs = [r[0] for r in cursor.fetchall()]
            if self.upload_config.catalog not in catalogs:
                raise ValueError(
                    "Catalog {} not found in {}".format(
                        self.upload_config.catalog, ", ".join(catalogs)
                    )
                )
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            cursor.execute("SHOW DATABASES")
            databases = [r[0] for r in cursor.fetchall()]
            if self.upload_config.database not in databases:
                raise ValueError(
                    "Database {} not found in {}".format(
                        self.upload_config.database, ", ".join(databases)
                    )
                )
            cursor.execute(f"SHOW TABLES IN {self.upload_config.database}")
            table_names = [r[1] for r in cursor.fetchall()]
            if self.upload_config.table_name not in table_names:
                raise ValueError(
                    "Table {} not found in {}".format(
                        self.upload_config.table_name, ", ".join(table_names)
                    )
                )

    def get_output_path(self, file_data: FileData, suffix: str = ".json") -> str:
        filename = Path(file_data.source_identifiers.filename)
        adjusted_filename = filename if filename.suffix == suffix else f"{filename}{suffix}"
        return os.path.join(self.upload_config.path, f"{adjusted_filename}")

    @contextmanager
    def get_cursor(self, **connect_kwargs) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor(**connect_kwargs) as cursor:
            logger.debug(f"executing: USE CATALOG: '{self.upload_config.catalog}'")
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            logger.debug(f"executing: USE DATABASE: {self.upload_config.database}")
            cursor.execute(f"USE DATABASE {self.upload_config.database}")
            yield cursor

    def get_table_columns(self) -> dict[str, str]:
        if self._columns is None:
            with self.get_cursor() as cursor:
                cursor.execute(f"SELECT * from {self.upload_config.table_name} LIMIT 1")
                self._columns = {desc[0]: desc[1] for desc in cursor.description}
        return self._columns

    def _fit_to_schema(self, df: "DataFrame", add_missing_columns: bool = True) -> "DataFrame":
        import pandas as pd

        table_columns = self.get_table_columns()
        columns = set(df.columns)
        schema_fields = set(table_columns.keys())
        columns_to_drop = columns - schema_fields
        missing_columns = schema_fields - columns

        if columns_to_drop:
            logger.info(
                "Following columns will be dropped to match the table's schema: "
                f"{', '.join(columns_to_drop)}"
            )
        if missing_columns and add_missing_columns:
            logger.info(
                "Following null filled columns will be added to match the table's schema:"
                f" {', '.join(missing_columns)} "
            )

        df = df.drop(columns=columns_to_drop)

        if add_missing_columns:
            for column in missing_columns:
                df[column] = pd.Series()
        return df

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            df = get_data_df()
            df = self._fit_to_schema(df=df)
            temp_path = Path(temp_dir) / path.name
            df.to_json(temp_path, orient="records", lines=False)
            with self.get_cursor(staging_allowed_local_path=temp_dir) as cursor:
                catalog_path = self.get_output_path(file_data=file_data)
                logger.debug(f"uploading {path.as_posix()} to {catalog_path}")
                cursor.execute(f"PUT '{temp_path.as_posix()}' INTO '{catalog_path}' OVERWRITE")
                logger.debug(
                    f"migrating content from {catalog_path} to "
                    f"table {self.upload_config.table_name}"
                )
                columns = list(df.columns)
                column_str = ", ".join(columns)
                sql_statment = f"INSERT INTO `{self.upload_config.table_name}` ({column_str}) SELECT {column_str} FROM json.`{catalog_path}`"  # noqa: E501
                cursor.execute(sql_statment)


databricks_volumes_delta_tables_destination_entry = DestinationRegistryEntry(
    connection_config=DatabricksDeltaTablesConnectionConfig,
    uploader=DatabricksVolumeDeltaTableUploader,
    uploader_config=DatabricksVolumeDeltaTableUploaderConfig,
    upload_stager=DatabricksVolumeDeltaTableStager,
    upload_stager_config=DatabricksDeltaTablesUploadStagerConfig,
)



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from unstructured_ingest.interfaces import (
    BaseConnectorConfig,
    BaseDestinationConnector,
    WriteConfig,
)


@dataclass
class Writer(ABC):
    connector_config: BaseConnectorConfig
    write_config: WriteConfig

    @abstractmethod
    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        pass

    def get_connector(self, **kwargs) -> BaseDestinationConnector:
        connector_cls = self.get_connector_cls()
        return connector_cls(
            write_config=self.write_config,
            connector_config=self.connector_config,
        )



from __future__ import annotations

import asyncio
import logging
import multiprocessing as mp
import shutil
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import Any

from unstructured_ingest.v2.interfaces import ProcessorConfig, Uploader
from unstructured_ingest.v2.logger import logger, make_default_logger
from unstructured_ingest.v2.otel import OtelHandler
from unstructured_ingest.v2.pipeline.interfaces import PipelineStep
from unstructured_ingest.v2.pipeline.steps.chunk import Chunker, ChunkStep
from unstructured_ingest.v2.pipeline.steps.download import DownloaderT, DownloadStep
from unstructured_ingest.v2.pipeline.steps.embed import Embedder, EmbedStep
from unstructured_ingest.v2.pipeline.steps.filter import Filterer, FilterStep
from unstructured_ingest.v2.pipeline.steps.index import IndexerT, IndexStep
from unstructured_ingest.v2.pipeline.steps.partition import Partitioner, PartitionStep
from unstructured_ingest.v2.pipeline.steps.stage import UploadStager, UploadStageStep
from unstructured_ingest.v2.pipeline.steps.uncompress import Uncompressor, UncompressStep
from unstructured_ingest.v2.pipeline.steps.upload import UploadStep
from unstructured_ingest.v2.processes.chunker import ChunkerConfig
from unstructured_ingest.v2.processes.connector_registry import (
    ConnectionConfig,
    DownloaderConfigT,
    IndexerConfigT,
    UploaderConfigT,
    UploadStagerConfigT,
    destination_registry,
    source_registry,
)
from unstructured_ingest.v2.processes.connectors.local import LocalUploader
from unstructured_ingest.v2.processes.embedder import EmbedderConfig
from unstructured_ingest.v2.processes.filter import FiltererConfig
from unstructured_ingest.v2.processes.partitioner import PartitionerConfig


class PipelineError(Exception):
    pass


@dataclass
class Pipeline:
    context: ProcessorConfig

    indexer: InitVar[IndexerT]
    indexer_step: IndexStep = field(init=False)

    downloader: InitVar[DownloaderT]
    downloader_step: DownloadStep = field(init=False)

    partitioner: InitVar[Partitioner]
    partitioner_step: PartitionStep = field(init=False)

    chunker: InitVar[Chunker | None] = None
    chunker_step: ChunkStep | None = field(init=False, default=None)

    embedder: InitVar[Embedder | None] = None
    embedder_step: EmbedStep | None = field(init=False, default=None)

    stager: InitVar[UploadStager | None] = None
    stager_step: UploadStageStep | None = field(init=False, default=None)

    uploader: InitVar[Uploader] = field(default=LocalUploader())
    uploader_step: UploadStep | None = field(init=False, default=None)

    uncompress_step: UncompressStep | None = field(init=False, default=None)

    filterer: InitVar[Filterer | None] = None
    filter_step: FilterStep | None = field(init=False, default=None)

    def __post_init__(
        self,
        indexer: IndexerT,
        downloader: DownloaderT,
        partitioner: Partitioner,
        chunker: Chunker | None = None,
        embedder: Embedder | None = None,
        stager: UploadStager | None = None,
        uploader: Uploader | None = None,
        filterer: Filterer | None = None,
    ):
        make_default_logger(level=logging.DEBUG if self.context.verbose else logging.INFO)
        otel_handler = OtelHandler(otel_endpoint=self.context.otel_endpoint)
        otel_handler.init_trace()
        self.indexer_step = IndexStep(process=indexer, context=self.context)
        self.downloader_step = DownloadStep(process=downloader, context=self.context)
        self.filter_step = FilterStep(process=filterer, context=self.context) if filterer else None
        self.partitioner_step = PartitionStep(process=partitioner, context=self.context)
        self.chunker_step = ChunkStep(process=chunker, context=self.context) if chunker else None

        self.embedder_step = EmbedStep(process=embedder, context=self.context) if embedder else None

        self.stager_step = UploadStageStep(process=stager, context=self.context) if stager else None
        self.uploader_step = UploadStep(process=uploader, context=self.context)
        if self.context.uncompress:
            process = Uncompressor()
            self.uncompress_step = UncompressStep(process=process, context=self.context)

        self.check_destination_connector()

    def check_destination_connector(self):
        # Make sure that if the set destination connector expects a stager, one is also set
        if not self.uploader_step:
            return
        uploader_connector_type = self.uploader_step.process.connector_type
        registry_entry = destination_registry[uploader_connector_type]
        if registry_entry.upload_stager and self.stager_step is None:
            try:
                self.stager_step = UploadStageStep(
                    process=registry_entry.upload_stager(), context=self.context
                )
                return
            except Exception as e:
                logger.debug(f"failed to instantiate required stager on user's behalf: {e}")
            raise ValueError(
                f"pipeline with uploader type {self.uploader_step.process.__class__.__name__} "
                f"expects a stager of type {registry_entry.upload_stager.__name__} "
                f"but one was not set"
            )

    def cleanup(self):
        if self.context.delete_cache and Path(self.context.work_dir).exists():
            logger.info(f"deleting cache directory: {self.context.work_dir}")
            shutil.rmtree(self.context.work_dir)

    def log_statuses(self):
        if status := self.context.status:
            logger.error(f"{len(status)} failed documents:")
            for k, v in status.items():
                for kk, vv in v.items():
                    logger.error(f"{k}: [{kk}] {vv}")

    def _run_initialization(self):
        failures = {}
        init_kwargs = {}
        for step in self._get_ordered_steps():
            try:
                step.process.init(**init_kwargs)
                step.process.precheck()
                # Make sure embedder dimensions available for downstream steps
                if isinstance(step.process, Embedder):
                    embed_dimensions = step.process.config.get_embedder().dimension
                    init_kwargs["vector_length"] = embed_dimensions

            except Exception as e:
                failures[step.process.__class__.__name__] = f"[{type(e).__name__}] {e}"
        if failures:
            for k, v in failures.items():
                logger.error(f"Step initialization failure: {k}: {v}")
            raise PipelineError("Initialization failed")

    def run(self):
        otel_handler = OtelHandler(otel_endpoint=self.context.otel_endpoint, log_out=logger.info)
        try:
            with otel_handler.get_tracer().start_as_current_span(
                "ingest process", record_exception=True
            ):
                self._run_initialization()
                self._run()
        finally:
            self.log_statuses()
            self.cleanup()
            if self.context.status:
                raise PipelineError("Pipeline did not run successfully")

    def clean_results(self, results: list[Any | list[Any]] | None) -> list[Any] | None:
        if not results:
            return None
        results = [r for r in results if r]
        flat = []
        for r in results:
            if isinstance(r, list):
                flat.extend(r)
            else:
                flat.append(r)
        final = [f for f in flat if f]
        return final or None

    def _get_ordered_steps(self) -> list[PipelineStep]:
        steps = [self.indexer_step, self.downloader_step]
        if self.uncompress_step:
            steps.append(self.uncompress_step)
        steps.append(self.partitioner_step)
        if self.chunker_step:
            steps.append(self.chunker_step)
        if self.embedder_step:
            steps.append(self.embedder_step)
        if self.stager_step:
            steps.append(self.stager_step)
        steps.append(self.uploader_step)
        return steps

    def apply_filter(self, records: list[dict]) -> list[dict]:
        if not self.filter_step:
            return records
        data_to_filter = [{"file_data_path": i["file_data_path"]} for i in records]
        filtered_data = self.filter_step(data_to_filter)
        filtered_data = [f for f in filtered_data if f is not None]
        filtered_file_data_paths = [r["file_data_path"] for r in filtered_data]
        filtered_records = [r for r in records if r["file_data_path"] in filtered_file_data_paths]
        return filtered_records

    def get_indices(self) -> list[dict]:
        if self.indexer_step.process.is_async():

            async def run_async():
                output = []
                async for i in self.indexer_step.run_async():
                    output.append(i)
                return output

            indices = asyncio.run(run_async())
        else:
            indices = self.indexer_step.run()
        indices_inputs = [{"file_data_path": i} for i in indices]
        return indices_inputs

    def _run(self):
        logger.info(
            f"running local pipeline: {self} with configs: " f"{self.context.model_dump_json()}"
        )
        if self.context.mp_supported:
            manager = mp.Manager()
            self.context.status = manager.dict()
        else:
            self.context.status = {}

        # Index into data source
        indices_inputs = self.get_indices()
        if not indices_inputs:
            logger.info("No files to process after indexer, exiting")
            return

        # Initial filtering on indexed content
        indices_inputs = self.apply_filter(records=indices_inputs)
        if not indices_inputs:
            logger.info("No files to process after filtering indexed content, exiting")
            return

        # Download associated content to local file system
        downloaded_data = self.downloader_step(indices_inputs)
        downloaded_data = self.clean_results(results=downloaded_data)
        if not downloaded_data:
            logger.info("No files to process after downloader, exiting")
            return

        # Post download filtering
        downloaded_data = self.apply_filter(records=downloaded_data)
        if not downloaded_data:
            logger.info("No files to process after filtering downloaded content, exiting")
            return

        # Run uncompress if available
        if self.uncompress_step:
            downloaded_data = self.uncompress_step(downloaded_data)
            # Flatten list of lists
            downloaded_data = self.clean_results(results=downloaded_data)

            # Post uncompress filtering
            downloaded_data = self.apply_filter(records=downloaded_data)
            if not downloaded_data:
                logger.info("No files to process after filtering uncompressed content, exiting")
                return

        if not downloaded_data or self.context.download_only:
            return

        # Partition content
        elements = self.partitioner_step(downloaded_data)
        elements = self.clean_results(results=elements)
        # Download data non longer needed, delete if possible
        self.downloader_step.delete_cache()
        elements = self.clean_results(results=elements)
        if not elements:
            logger.info("No files to process after partitioning, exiting")
            return

        # Run element specific modifiers
        last_step = self.partitioner_step
        for step in [s for s in [self.chunker_step, self.embedder_step, self.stager_step] if s]:
            elements = step(elements)
            elements = self.clean_results(results=elements)
            # Delete data from previous step if possible since no longer needed
            last_step.delete_cache()
            last_step = step
            if not elements:
                logger.info(f"no files to process after {step.__class__.__name__}, exiting")
                return

        # Upload the final result
        self.uploader_step(iterable=elements)
        last_step.delete_cache()

    def __str__(self):
        s = [str(self.indexer_step)]
        if filter_step := self.filter_step:
            s.append(str(filter_step))
        s.append(str(self.downloader_step))
        if filter_step := self.filter_step:
            s.append(str(filter_step))
        if uncompress_step := self.uncompress_step:
            s.extend([str(uncompress_step), str(filter_step)])
        s.append(str(self.partitioner_step))
        if chunker_step := self.chunker_step:
            s.append(str(chunker_step))
        if embedder_step := self.embedder_step:
            s.append(str(embedder_step))
        if stager_step := self.stager_step:
            s.append(str(stager_step))
        s.append(str(self.uploader_step))
        return " -> ".join(s)

    @classmethod
    def from_configs(
        cls,
        context: ProcessorConfig,
        indexer_config: IndexerConfigT,
        downloader_config: DownloaderConfigT,
        source_connection_config: ConnectionConfig,
        partitioner_config: PartitionerConfig,
        filterer_config: FiltererConfig | None = None,
        chunker_config: ChunkerConfig | None = None,
        embedder_config: EmbedderConfig | None = None,
        destination_connection_config: ConnectionConfig | None = None,
        stager_config: UploadStagerConfigT | None = None,
        uploader_config: UploaderConfigT | None = None,
    ) -> "Pipeline":
        # Get registry key based on indexer config
        source_entry = {
            k: v
            for k, v in source_registry.items()
            if type(indexer_config) is v.indexer_config
            and type(downloader_config) is v.downloader_config
            and type(source_connection_config) is v.connection_config
        }
        if len(source_entry) > 1:
            raise ValueError(
                f"multiple entries found matching provided indexer, "
                f"downloader and connection configs: {source_entry}"
            )
        if len(source_entry) != 1:
            raise ValueError(
                "no entry found in source registry with matching indexer, "
                "downloader and connection configs"
            )
        source = list(source_entry.values())[0]
        pipeline_kwargs = {
            "context": context,
            "indexer": source.indexer(
                index_config=indexer_config, connection_config=source_connection_config
            ),
            "downloader": source.downloader(
                download_config=downloader_config, connection_config=source_connection_config
            ),
            "partitioner": Partitioner(config=partitioner_config),
        }
        if filterer_config:
            pipeline_kwargs["filterer"] = Filterer(config=filterer_config)
        if chunker_config:
            pipeline_kwargs["chunker"] = Chunker(config=chunker_config)
        if embedder_config:
            pipeline_kwargs["embedder"] = Embedder(config=embedder_config)
        if not uploader_config:
            return Pipeline(**pipeline_kwargs)

        destination_entry = {
            k: v
            for k, v in destination_registry.items()
            if isinstance(uploader_config, v.uploader_config)
        }
        if destination_connection_config:
            destination_entry = {
                k: v
                for k, v in destination_entry.items()
                if isinstance(destination_connection_config, v.connection_config)
            }
        if stager_config:
            destination_entry = {
                k: v
                for k, v in destination_entry.items()
                if isinstance(stager_config, v.upload_stager_config)
            }

        if len(destination_entry) > 1:
            raise ValueError(
                f"multiple entries found matching provided uploader, "
                f"stager and connection configs: {destination_entry}"
            )
        if len(destination_entry) != 1:
            raise ValueError(
                "no entry found in destination registry with matching uploader, "
                "stager and connection configs"
            )

        destination = list(destination_entry.values())[0]
        if stager_config:
            pipeline_kwargs["stager"] = destination.upload_stager(
                upload_stager_config=stager_config
            )
        if uploader_config:
            uploader_kwargs = {"upload_config": uploader_config}
            if destination_connection_config:
                uploader_kwargs["connection_config"] = destination_connection_config
            pipeline_kwargs["uploader"] = destination.uploader(**uploader_kwargs)
        return cls(**pipeline_kwargs)



import os
import shutil
from pathlib import Path

from test.integration.connectors.utils.validation.utils import ValidationConfig
from unstructured_ingest.utils.data_prep import get_data
from unstructured_ingest.v2.interfaces import UploadStager
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


class StagerValidationConfigs(ValidationConfig):
    expected_count: int
    expected_folder: str = "stager"

    def stager_output_dir(self) -> Path:
        dir = self.test_output_dir() / self.expected_folder
        dir.mkdir(exist_ok=True, parents=True)
        return dir

    def stager_output_path(self, input_path: Path) -> Path:
        return self.stager_output_dir() / input_path.name


def run_all_stager_validations(
    configs: StagerValidationConfigs, input_file: Path, staged_filepath: Path
):
    # Validate matching extensions
    assert input_file.suffix == staged_filepath.suffix

    # Validate length
    staged_data = get_data(path=staged_filepath)
    assert len(staged_data) == configs.expected_count

    # Validate file
    expected_filepath = configs.stager_output_path(input_path=input_file)
    assert expected_filepath.exists(), f"{expected_filepath} does not exist"
    assert expected_filepath.is_file(), f"{expected_filepath} is not a file"
    if configs.detect_diff(expected_filepath=expected_filepath, current_filepath=staged_filepath):
        raise AssertionError(
            f"Current file ({staged_filepath}) does not match expected file: {expected_filepath}"
        )


def update_stager_fixtures(stager_output_path: Path, staged_filepath: Path):
    copied_filepath = stager_output_path / staged_filepath.name
    shutil.copy(staged_filepath, copied_filepath)


def stager_validation(
    stager: UploadStager,
    tmp_dir: Path,
    input_file: Path,
    configs: StagerValidationConfigs,
    overwrite_fixtures: bool = os.getenv("OVERWRITE_FIXTURES", "False").lower() == "true",
) -> None:
    # Run stager
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=input_file.name, filename=input_file.name),
        connector_type=configs.test_id,
        identifier="mock file data",
    )
    staged_filepath = stager.run(
        elements_filepath=input_file,
        file_data=file_data,
        output_dir=tmp_dir,
        output_filename=input_file.name,
    )
    if not overwrite_fixtures:
        print("Running validation")
        run_all_stager_validations(
            configs=configs, input_file=input_file, staged_filepath=staged_filepath
        )
    else:
        print("Running fixtures update")
        update_stager_fixtures(
            stager_output_path=configs.stager_output_dir(), staged_filepath=staged_filepath
        )



import asyncio
import json
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, AsyncGenerator, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.error import DestinationConnectionError, WriteError
from unstructured_ingest.utils.data_prep import batch_generator, flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.types.file_data import FileData
from unstructured_ingest.v2.utils import get_enhanced_element_id

if TYPE_CHECKING:
    from qdrant_client import AsyncQdrantClient, QdrantClient


class QdrantAccessConfig(AccessConfig, ABC):
    pass


class QdrantConnectionConfig(ConnectionConfig, ABC):
    access_config: Secret[QdrantAccessConfig] = Field(
        default_factory=QdrantAccessConfig, validate_default=True, description="Access Config"
    )

    @abstractmethod
    def get_client_kwargs(self) -> dict:
        pass

    @requires_dependencies(["qdrant_client"], extras="qdrant")
    @asynccontextmanager
    async def get_async_client(self) -> AsyncGenerator["AsyncQdrantClient", None]:
        from qdrant_client import AsyncQdrantClient

        client_kwargs = self.get_client_kwargs()
        client = AsyncQdrantClient(**client_kwargs)
        try:
            yield client
        finally:
            await client.close()

    @requires_dependencies(["qdrant_client"], extras="qdrant")
    @contextmanager
    def get_client(self) -> Generator["QdrantClient", None, None]:
        from qdrant_client import QdrantClient

        client_kwargs = self.get_client_kwargs()
        client = QdrantClient(**client_kwargs)
        try:
            yield client
        finally:
            client.close()


class QdrantUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class QdrantUploadStager(UploadStager, ABC):
    upload_stager_config: QdrantUploadStagerConfig = field(
        default_factory=lambda: QdrantUploadStagerConfig()
    )

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        """Prepares dictionary in the format that Chroma requires"""
        data = element_dict.copy()
        return {
            "id": get_enhanced_element_id(element_dict=data, file_data=file_data),
            "vector": data.pop("embeddings", {}),
            "payload": {
                "text": data.pop("text", None),
                "element_serialized": json.dumps(data),
                **flatten_dict(
                    data,
                    separator="-",
                    flatten_lists=True,
                ),
            },
        }


class QdrantUploaderConfig(UploaderConfig):
    collection_name: str = Field(description="Name of the collection.")
    batch_size: int = Field(default=50, description="Number of records per batch.")
    num_processes: Optional[int] = Field(
        default=1,
        description="Optional limit on number of threads to use for upload.",
        deprecated=True,
    )


@dataclass
class QdrantUploader(Uploader, ABC):
    upload_config: QdrantUploaderConfig
    connection_config: QdrantConnectionConfig

    @DestinationConnectionError.wrap
    def precheck(self) -> None:
        with self.connection_config.get_client() as client:
            collections_response = client.get_collections()
            collection_names = [c.name for c in collections_response.collections]
            if self.upload_config.collection_name not in collection_names:
                raise DestinationConnectionError(
                    "collection '{}' not found: {}".format(
                        self.upload_config.collection_name, ", ".join(collection_names)
                    )
                )

    def is_async(self):
        return True

    async def run_data_async(
        self,
        data: list[dict],
        file_data: FileData,
        **kwargs: Any,
    ) -> None:
        batches = list(batch_generator(data, batch_size=self.upload_config.batch_size))
        logger.debug(
            "Elements split into %i batches of size %i.",
            len(batches),
            self.upload_config.batch_size,
        )
        await asyncio.gather(*[self._upsert_batch(batch) for batch in batches])

    async def _upsert_batch(self, batch: list[dict]) -> None:
        from qdrant_client import models

        points: list[models.PointStruct] = [models.PointStruct(**item) for item in batch]
        try:
            logger.debug(
                "Upserting %i points to the '%s' collection.",
                len(points),
                self.upload_config.collection_name,
            )
            async with self.connection_config.get_async_client() as async_client:
                await async_client.upsert(
                    self.upload_config.collection_name, points=points, wait=True
                )
        except Exception as api_error:
            logger.error(
                "Failed to upsert points to the collection due to the following error %s", api_error
            )

            raise WriteError(f"Qdrant error: {api_error}") from api_error

        logger.debug("Successfully upsert points to the collection.")



from dataclasses import dataclass
from time import time
from typing import TYPE_CHECKING, Any, AsyncGenerator, Generator, Optional

from pydantic import UUID4, Field, Secret

from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import SourceRegistryEntry
from unstructured_ingest.v2.types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)

if TYPE_CHECKING:
    from unstructured_ingest.v2.processes.connectors.notion.client import Client

NOTION_API_VERSION = "2022-06-28"
CONNECTOR_TYPE = "notion"


class NotionAccessConfig(AccessConfig):
    notion_api_key: str = Field(description="Notion API key")


class NotionConnectionConfig(ConnectionConfig):
    access_config: Secret[NotionAccessConfig]

    @requires_dependencies(["notion_client"], extras="notion")
    def get_client(self) -> "Client":
        from unstructured_ingest.v2.processes.connectors.notion.client import Client

        return Client(
            notion_version=NOTION_API_VERSION,
            auth=self.access_config.get_secret_value().notion_api_key,
            logger=logger,
            log_level=logger.level,
        )


class NotionIndexerConfig(IndexerConfig):
    page_ids: Optional[list[str]] = Field(
        default=None, description="List of Notion page IDs to process"
    )

    database_ids: Optional[list[str]] = Field(
        default=None, description="List of Notion database IDs to process"
    )
    recursive: bool = Field(
        default=False, description="Recursively process child pages and databases"
    )

    def __post_init__(self):
        if self.page_ids:
            self.page_ids: list[UUID4] = [UUID4(p.strip()) for p in self.page_ids]

        if self.database_ids:
            self.database_ids: list[UUID4] = [UUID4(p.strip()) for p in self.database_ids]


@dataclass
class NotionIndexer(Indexer):
    connection_config: NotionConnectionConfig
    index_config: NotionIndexerConfig

    def is_async(self) -> bool:
        return False

    def precheck(self) -> None:
        """Check the connection to the Notion API."""
        try:
            client = self.connection_config.get_client()
            # Perform a simple request to verify connection
            request = client._build_request("HEAD", "users")
            response = client.client.send(request)
            response.raise_for_status()

        except Exception as e:
            logger.error(f"Failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to validate connection: {e}")

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        client = self.connection_config.get_client()
        processed_pages: set[str] = set()
        processed_databases: set[str] = set()

        pages_to_process: set[str] = set(self.index_config.page_ids or [])
        databases_to_process: set[str] = set(self.index_config.database_ids or [])

        while pages_to_process or databases_to_process:
            # Process pages
            for page_id in list(pages_to_process):
                if page_id in processed_pages:
                    continue

                processed_pages.add(page_id)
                pages_to_process.remove(page_id)
                file_data = self.get_page_file_data(page_id=page_id, client=client)
                if file_data:
                    yield file_data

                if self.index_config.recursive:
                    (child_pages, child_databases) = self.get_child_pages_and_databases(
                        page_id=page_id,
                        client=client,
                        processed_pages=processed_pages,
                        processed_databases=processed_databases,
                    )
                    pages_to_process.update(child_pages)
                    databases_to_process.update(child_databases)

            # Process databases
            for database_id in list(databases_to_process):
                if database_id in processed_databases:
                    continue
                processed_databases.add(database_id)
                databases_to_process.remove(database_id)
                file_data = self.get_database_file_data(database_id=database_id, client=client)
                if file_data:
                    yield file_data
                if self.index_config.recursive:
                    (
                        child_pages,
                        child_databases,
                    ) = self.get_child_pages_and_databases_from_database(
                        database_id=database_id,
                        client=client,
                        processed_pages=processed_pages,
                        processed_databases=processed_databases,
                    )
                    pages_to_process.update(child_pages)
                    databases_to_process.update(child_databases)

    @requires_dependencies(["notion_client"], extras="notion")
    def get_page_file_data(self, page_id: str, client: "Client") -> Optional[FileData]:
        try:
            page_metadata = client.pages.retrieve(page_id=page_id)  # type: ignore
            date_created = page_metadata.created_time
            date_modified = page_metadata.last_edited_time
            identifier = page_id
            source_identifiers = SourceIdentifiers(
                filename=f"{page_id}.html",
                fullpath=f"{page_id}.html",
                rel_path=f"{page_id}.html",
            )
            metadata = FileDataSourceMetadata(
                date_created=date_created,
                date_modified=date_modified,
                record_locator={"page_id": page_id},
                date_processed=str(time()),
            )
            # additional_metadata = page_metadata
            additional_metadata = {
                "created_by": page_metadata.created_by,
                "last_edited_by": page_metadata.last_edited_by,
                "parent": page_metadata.parent,
                "url": page_metadata.url,
            }

            return FileData(
                identifier=identifier,
                connector_type=CONNECTOR_TYPE,
                source_identifiers=source_identifiers,
                metadata=metadata,
                additional_metadata=additional_metadata,
            )
        except Exception as e:
            logger.error(f"Error retrieving page {page_id}: {e}")
            return None

    @requires_dependencies(["notion_client"], extras="notion")
    def get_database_file_data(self, database_id: str, client: "Client") -> Optional[FileData]:
        try:
            # type: ignore
            database_metadata = client.databases.retrieve(database_id=database_id)
            date_created = database_metadata.created_time
            date_modified = database_metadata.last_edited_time
            identifier = database_id
            source_identifiers = SourceIdentifiers(
                filename=f"{database_id}.html",
                fullpath=f"{database_id}.html",
                rel_path=f"{database_id}.html",
            )
            metadata = FileDataSourceMetadata(
                date_created=date_created,
                date_modified=date_modified,
                record_locator={"database_id": database_id},
                date_processed=str(time()),
            )
            additional_metadata = {
                "created_by": database_metadata.created_by,
                "last_edited_by": database_metadata.last_edited_by,
                "parent": database_metadata.parent,
                "url": database_metadata.url,
            }
            return FileData(
                identifier=identifier,
                connector_type=CONNECTOR_TYPE,
                source_identifiers=source_identifiers,
                metadata=metadata,
                additional_metadata=additional_metadata,
            )
        except Exception as e:
            logger.error(f"Error retrieving database {database_id}: {e}")
            return None

    def get_child_pages_and_databases(
        self,
        page_id: str,
        client: "Client",
        processed_pages: set[str],
        processed_databases: set[str],
    ) -> tuple[set[str], set[str]]:
        from unstructured_ingest.v2.processes.connectors.notion.helpers import (
            get_recursive_content_from_page,
        )

        child_content = get_recursive_content_from_page(
            client=client,
            page_id=page_id,
            logger=logger,
        )
        child_pages = set(child_content.child_pages) - processed_pages
        child_databases = set(child_content.child_databases) - processed_databases
        return child_pages, child_databases

    def get_child_pages_and_databases_from_database(
        self,
        database_id: str,
        client: "Client",
        processed_pages: set[str],
        processed_databases: set[str],
    ) -> tuple[set[str], set[str]]:
        from unstructured_ingest.v2.processes.connectors.notion.helpers import (
            get_recursive_content_from_database,
        )

        child_content = get_recursive_content_from_database(
            client=client,
            database_id=database_id,
            logger=logger,
        )
        child_pages = set(child_content.child_pages) - processed_pages
        child_databases = set(child_content.child_databases) - processed_databases
        return child_pages, child_databases

    async def run_async(self, **kwargs: Any) -> AsyncGenerator[None, None]:
        # Asynchronous run is not implemented
        raise NotImplementedError()


class NotionDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class NotionDownloader(Downloader):
    connection_config: NotionConnectionConfig
    download_config: NotionDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        client = self.connection_config.get_client()
        record_locator = file_data.metadata.record_locator

        if "page_id" in record_locator:
            return self.download_page(
                client=client,
                page_id=record_locator["page_id"],
                file_data=file_data,
            )
        elif "database_id" in record_locator:
            return self.download_database(
                client=client,
                database_id=record_locator["database_id"],
                file_data=file_data,
            )
        else:
            raise ValueError("Invalid record_locator in file_data")

    def download_page(self, client, page_id: str, file_data: FileData) -> DownloadResponse:
        from unstructured_ingest.v2.processes.connectors.notion.helpers import extract_page_html

        try:
            text_extraction = extract_page_html(
                client=client,
                page_id=page_id,
                logger=logger,
            )

            if text_extraction.html:
                download_path = self.get_download_path(file_data=file_data)
                download_path.parent.mkdir(parents=True, exist_ok=True)
                with download_path.open("w") as page_file:
                    page_file.write(text_extraction.html.render(pretty=True))
                return self.generate_download_response(
                    file_data=file_data, download_path=download_path
                )
            else:
                logger.error(f"No HTML content for page {page_id}")
                return None
        except Exception as e:
            logger.error(f"Error downloading page {page_id}: {e}")
            return None

    def download_database(self, client, database_id: str, file_data: FileData) -> DownloadResponse:
        from unstructured_ingest.v2.processes.connectors.notion.helpers import extract_database_html

        try:
            text_extraction = extract_database_html(
                client=client,
                database_id=database_id,
                logger=logger,
            )
            if text_extraction.html:
                download_path = self.get_download_path(file_data=file_data)
                download_path.parent.mkdir(parents=True, exist_ok=True)
                with download_path.open("w") as database_file:
                    database_file.write(text_extraction.html.render(pretty=True))
                return self.generate_download_response(
                    file_data=file_data, download_path=download_path
                )
            else:
                logger.error(f"No HTML content for database {database_id}")
                return None
        except Exception as e:
            logger.error(f"Error downloading database {database_id}: {e}")
            return None


notion_source_entry = SourceRegistryEntry(
    connection_config=NotionConnectionConfig,
    indexer_config=NotionIndexerConfig,
    indexer=NotionIndexer,
    downloader_config=NotionDownloaderConfig,
    downloader=NotionDownloader,
)



import pytest
from pydantic import ValidationError

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from unstructured_ingest.v2.processes.connectors.weaviate.cloud import (
    CONNECTOR_TYPE,
    CloudWeaviateAccessConfig,
    CloudWeaviateConnectionConfig,
)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_weaviate_failing_connection_config():
    with pytest.raises(ValidationError):
        CloudWeaviateConnectionConfig(
            access_config=CloudWeaviateAccessConfig(api_key="my key", password="password"),
            username="username",
            cluster_url="clusterurl",
        )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_weaviate_connection_config_happy_path():
    CloudWeaviateConnectionConfig(
        access_config=CloudWeaviateAccessConfig(
            api_key="my key",
        ),
        cluster_url="clusterurl",
    )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_weaviate_connection_config_anonymous():
    CloudWeaviateConnectionConfig(
        access_config=CloudWeaviateAccessConfig(api_key="my key", password="password"),
        username="username",
        anonymous=True,
        cluster_url="clusterurl",
    )



# https://developers.notion.com/reference/property-object#multi-select
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.attributes import Style
from htmlBuilder.tags import Div, HtmlTag, Span

from unstructured_ingest.connector.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class MultiSelectOption(FromJSONMixin):
    color: str
    id: str
    name: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class MultiSelectProp(FromJSONMixin):
    options: List[MultiSelectOption] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(options=[MultiSelectOption.from_dict(o) for o in data.get("options", [])])


@dataclass
class MultiSelect(DBPropertyBase):
    id: str
    name: str
    multi_select: MultiSelectProp
    type: str = "multi_select"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            multi_select=data.pop("multi_select", {}),
            **data,
        )


@dataclass
class MultiSelectCell(DBCellBase):
    id: str
    multi_select: List[MultiSelectOption]
    type: str = "multi_select"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            multi_select=[MultiSelectOption.from_dict(o) for o in data.pop("multi_select", [])],
            **data,
        )

    def get_html(self) -> Optional[HtmlTag]:
        if not self.multi_select:
            return None
        option_spans = []
        for option in self.multi_select:
            option_attributes = []
            if option.color and option.color != "default":
                option_attributes.append(Style(f"color: {option.color}"))
            option_spans.append(Span(option_attributes, option.name))
        return Div([], option_spans)



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
    CliRecursiveConfig,
    DelimitedString,
)
from unstructured_ingest.connector.local import SimpleLocalConfig


@dataclass
class LocalCliConfig(SimpleLocalConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--input-path"],
                required=True,
                type=click.Path(file_okay=True, dir_okay=True, exists=True),
                help="Path to the location in the local file system that will be processed.",
            ),
            click.Option(
                ["--file-glob"],
                default=None,
                type=DelimitedString(),
                help="A comma-separated list of file globs to limit which types of "
                "local files are accepted, e.g. '*.html,*.txt'",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="local",
        cli_config=LocalCliConfig,
        additional_cli_options=[CliRecursiveConfig],
    )
    return cmd_cls



# https://developers.notion.com/reference/property-object#last-edited-by
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.tags import HtmlTag

from unstructured_ingest.connector.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.connector.notion.types.user import People


@dataclass
class LastEditedBy(DBPropertyBase):
    @classmethod
    def from_dict(cls, data: dict):
        return cls()

    def get_text(self) -> Optional[str]:
        return None


@dataclass
class LastEditedByCell(DBCellBase):
    id: str
    last_edited_by: People
    type: str = "last_edited_by"

    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(last_edited_by=People.from_dict(data.pop("last_edited_by", {})), **data)

    def get_html(self) -> Optional[HtmlTag]:
        return self.last_edited_by.get_html()



from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from unstructured_ingest.v2.interfaces.upload_stager import UploadStagerConfig
from unstructured_ingest.v2.processes.connectors.duckdb.base import BaseDuckDBUploadStager
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@pytest.fixture
def mock_instance() -> BaseDuckDBUploadStager:
    return BaseDuckDBUploadStager(UploadStagerConfig())


@pytest.mark.parametrize(
    ("input_filepath", "output_filename", "expected"),
    [
        (
            "/path/to/input_file.ndjson",
            "output_file.ndjson",
            "output_file.ndjson",
        ),
        ("input_file.txt", "output_file.json", "output_file.txt"),
        ("/path/to/input_file.json", "output_file", "output_file.json"),
    ],
)
def test_run_output_filename_suffix(
    mocker: MockerFixture,
    mock_instance: BaseDuckDBUploadStager,
    input_filepath: str,
    output_filename: str,
    expected: str,
):
    output_dir = Path("/tmp/test/output_dir")

    # Mocks
    mock_get_data = mocker.patch(
        "unstructured_ingest.v2.processes.connectors.duckdb.base.get_data",
        return_value=[{"key": "value"}, {"key": "value2"}],
    )
    mock_conform_dict = mocker.patch.object(
        BaseDuckDBUploadStager,
        "conform_dict",
        side_effect=lambda element_dict, file_data: element_dict,
    )
    mock_get_output_path = mocker.patch.object(
        BaseDuckDBUploadStager, "get_output_path", return_value=output_dir / expected
    )
    mock_write_output = mocker.patch(
        "unstructured_ingest.v2.processes.connectors.duckdb.base.write_data", return_value=None
    )

    # Act
    result = mock_instance.run(
        elements_filepath=Path(input_filepath),
        file_data=FileData(
            identifier="test",
            connector_type="test",
            source_identifiers=SourceIdentifiers(filename=input_filepath, fullpath=input_filepath),
        ),
        output_dir=output_dir,
        output_filename=output_filename,
    )

    # Assert
    mock_get_data.assert_called_once_with(path=Path(input_filepath))
    assert mock_conform_dict.call_count == 2
    mock_get_output_path.assert_called_once_with(output_filename=expected, output_dir=output_dir)
    mock_write_output.assert_called_once_with(
        path=output_dir / expected, data=[{"key": "value"}, {"key": "value2"}]
    )
    assert result.name == expected



import json
import os
import uuid
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from unittest import mock

import pytest
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors.platform import NotFound

from test.integration.connectors.utils.constants import (
    BLOB_STORAGE_TAG,
    DESTINATION_TAG,
    SOURCE_TAG,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.errors import UserAuthError, UserError
from unstructured_ingest.v2.processes.connectors.databricks.volumes_native import (
    CONNECTOR_TYPE,
    DatabricksNativeVolumesAccessConfig,
    DatabricksNativeVolumesConnectionConfig,
    DatabricksNativeVolumesDownloader,
    DatabricksNativeVolumesDownloaderConfig,
    DatabricksNativeVolumesIndexer,
    DatabricksNativeVolumesIndexerConfig,
    DatabricksNativeVolumesUploader,
    DatabricksNativeVolumesUploaderConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@dataclass
class BaseEnvData:
    host: str
    catalog: str


@dataclass
class BasicAuthEnvData(BaseEnvData):
    client_id: str
    client_secret: str

    def get_connection_config(self) -> DatabricksNativeVolumesConnectionConfig:
        return DatabricksNativeVolumesConnectionConfig(
            host=self.host,
            access_config=DatabricksNativeVolumesAccessConfig(
                client_id=self.client_id,
                client_secret=self.client_secret,
            ),
        )


@dataclass
class PATEnvData(BaseEnvData):
    token: str

    def get_connection_config(self) -> DatabricksNativeVolumesConnectionConfig:
        return DatabricksNativeVolumesConnectionConfig(
            host=self.host,
            access_config=DatabricksNativeVolumesAccessConfig(
                token=self.token,
            ),
        )


def get_basic_auth_env_data() -> BasicAuthEnvData:
    return BasicAuthEnvData(
        host=os.environ["DATABRICKS_HOST"],
        client_id=os.environ["DATABRICKS_CLIENT_ID"],
        client_secret=os.environ["DATABRICKS_CLIENT_SECRET"],
        catalog=os.environ["DATABRICKS_CATALOG"],
    )


def get_pat_env_data() -> PATEnvData:
    return PATEnvData(
        host=os.environ["DATABRICKS_HOST"],
        catalog=os.environ["DATABRICKS_CATALOG"],
        token=os.environ["DATABRICKS_PAT"],
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env(
    "DATABRICKS_HOST", "DATABRICKS_CLIENT_ID", "DATABRICKS_CLIENT_SECRET", "DATABRICKS_CATALOG"
)
async def test_volumes_native_source(tmp_path: Path):
    env_data = get_basic_auth_env_data()
    with mock.patch.dict(os.environ, clear=True):
        indexer_config = DatabricksNativeVolumesIndexerConfig(
            recursive=True,
            volume="test-platform",
            volume_path="databricks-volumes-test-input",
            catalog=env_data.catalog,
        )
        connection_config = env_data.get_connection_config()
        download_config = DatabricksNativeVolumesDownloaderConfig(download_dir=tmp_path)
        indexer = DatabricksNativeVolumesIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        downloader = DatabricksNativeVolumesDownloader(
            connection_config=connection_config, download_config=download_config
        )
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=SourceValidationConfigs(
                test_id="databricks_volumes_native",
                expected_num_files=1,
            ),
        )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("DATABRICKS_HOST", "DATABRICKS_PAT", "DATABRICKS_CATALOG")
async def test_volumes_native_source_pat(tmp_path: Path):
    env_data = get_pat_env_data()
    with mock.patch.dict(os.environ, clear=True):
        indexer_config = DatabricksNativeVolumesIndexerConfig(
            recursive=True,
            volume="test-platform",
            volume_path="databricks-volumes-test-input",
            catalog=env_data.catalog,
        )
        connection_config = env_data.get_connection_config()
        download_config = DatabricksNativeVolumesDownloaderConfig(download_dir=tmp_path)
        indexer = DatabricksNativeVolumesIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        downloader = DatabricksNativeVolumesDownloader(
            connection_config=connection_config, download_config=download_config
        )
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=SourceValidationConfigs(
                test_id="databricks_volumes_native_pat",
                expected_num_files=1,
            ),
        )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("DATABRICKS_HOST", "DATABRICKS_PAT", "DATABRICKS_CATALOG")
def test_volumes_native_source_pat_invalid_catalog():
    env_data = get_pat_env_data()
    with mock.patch.dict(os.environ, clear=True):
        indexer_config = DatabricksNativeVolumesIndexerConfig(
            recursive=True,
            volume="test-platform",
            volume_path="databricks-volumes-test-input",
            catalog="fake_catalog",
        )
        indexer = DatabricksNativeVolumesIndexer(
            connection_config=env_data.get_connection_config(), index_config=indexer_config
        )
        with pytest.raises(UserError):
            _ = list(indexer.run())


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("DATABRICKS_HOST")
def test_volumes_native_source_pat_invalid_pat():
    host = os.environ["DATABRICKS_HOST"]
    with mock.patch.dict(os.environ, clear=True):
        indexer_config = DatabricksNativeVolumesIndexerConfig(
            recursive=True,
            volume="test-platform",
            volume_path="databricks-volumes-test-input",
            catalog="fake_catalog",
        )
        connection_config = DatabricksNativeVolumesConnectionConfig(
            host=host,
            access_config=DatabricksNativeVolumesAccessConfig(
                token="invalid-token",
            ),
        )
        indexer = DatabricksNativeVolumesIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        with pytest.raises(UserAuthError):
            _ = list(indexer.run())


def _get_volume_path(catalog: str, volume: str, volume_path: str):
    return f"/Volumes/{catalog}/default/{volume}/{volume_path}"


@contextmanager
def databricks_destination_context(
    env_data: BasicAuthEnvData, volume: str, volume_path
) -> WorkspaceClient:
    client = WorkspaceClient(
        host=env_data.host, client_id=env_data.client_id, client_secret=env_data.client_secret
    )
    try:
        yield client
    finally:
        # Cleanup
        try:
            for file in client.files.list_directory_contents(
                directory_path=_get_volume_path(env_data.catalog, volume, volume_path)
            ):
                client.files.delete(file.path)
            client.files.delete_directory(_get_volume_path(env_data.catalog, volume, volume_path))
        except NotFound:
            # Directory was never created, don't need to delete
            pass


def validate_upload(client: WorkspaceClient, catalog: str, volume: str, volume_path: str):
    files = list(
        client.files.list_directory_contents(
            directory_path=_get_volume_path(catalog, volume, volume_path)
        )
    )

    assert len(files) == 1

    resp = client.files.download(files[0].path)
    data = json.loads(resp.contents.read())

    assert len(data) == 22
    element_types = {v["type"] for v in data}
    assert len(element_types) == 1
    assert "CompositeElement" in element_types


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, BLOB_STORAGE_TAG)
@requires_env(
    "DATABRICKS_HOST", "DATABRICKS_CLIENT_ID", "DATABRICKS_CLIENT_SECRET", "DATABRICKS_CATALOG"
)
async def test_volumes_native_destination(upload_file: Path):
    env_data = get_basic_auth_env_data()
    volume_path = f"databricks-volumes-test-output-{uuid.uuid4()}"
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    with databricks_destination_context(
        volume="test-platform", volume_path=volume_path, env_data=env_data
    ) as workspace_client:
        connection_config = env_data.get_connection_config()
        uploader = DatabricksNativeVolumesUploader(
            connection_config=connection_config,
            upload_config=DatabricksNativeVolumesUploaderConfig(
                volume="test-platform",
                volume_path=volume_path,
                catalog=env_data.catalog,
            ),
        )
        uploader.precheck()
        if uploader.is_async():
            await uploader.run_async(path=upload_file, file_data=file_data)
        else:
            uploader.run(path=upload_file, file_data=file_data)

        validate_upload(
            client=workspace_client,
            catalog=env_data.catalog,
            volume="test-platform",
            volume_path=volume_path,
        )
